#! /usr/bin/env python

import gtk
import gobject
import pango
import sys

import os
import gc

BORDER_SIZE = 6
CELL_BORDER_WIDTH = 10
SELECTION_THICKNESS = 5
path_base = '/home/fred/multimedia/images/mine/photos/unsorted/2005-12-20--14.29.22/'

from libs import exif


class ThumbnailsView(gtk.Layout):
    scroll_value = 0
    scroll = False

    real_focus_cell = 0

    # initial thumbnail size
    real_thumbnail_width = 160
    thumbnail_height = 120

    def __init__(self):
        gtk.Layout.__init__(self, None, None)
        self.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse('white'))
        self.loading_pixbuf = gtk.gdk.pixbuf_new_from_file('gfx/refresh.png')
        self.selection = []
        self.filenames = []
        for filename in os.listdir(path_base):
            if not os.path.splitext(filename)[1].lower() in ('.jpg', '.jpeg'):
                continue
            self.filenames.append(os.path.join(path_base, filename))

        self.thumbnail_pixbufs = {}
        self.exif_tags = {}

        self.connect('expose-event', self.on_expose_event)
        self.connect('size-allocate', self.on_size_allocate)
        self.connect('set-scroll-adjustments', self.on_set_scroll_adjustments)
        self.connect('key-press-event', self.on_key_press)
        self.connect('button-press-event', self.on_button_press)
        self.set_flags(gtk.CAN_FOCUS)

        self.priority_loads = []
        def load_thumbnails_bg():
            for filename in self.filenames:
                for filename2 in self.priority_loads:
                    self.load_thumbnail(filename2)
                    self.invalidate_cell(self.filenames.index(filename2))
                    self.priority_loads = self.priority_loads[1:]
                    yield True
                if self.thumbnail_pixbufs.has_key(filename):
                    continue
                #print 'preloading', filename
                self.load_thumbnail(filename)
                self.invalidate_cell(self.filenames.index(filename))
                yield True
            yield False
        gobject.idle_add(load_thumbnails_bg().next)


    def get_thumbnail_width(self):
        return self.real_thumbnail_width
    def set_thumbnail_width(self, value):
        self.thumbnail_height = int(value * 3./4)
        self.real_thumbnail_width = int(value)
        if self.bin_window:
            rect = apply(gtk.gdk.Rectangle, tuple(self.allocation))
            rect.height = self.get_vadjustment().upper
            self.bin_window.invalidate_rect(rect, False)
        self.update_layout()
    thumbnail_width = property(get_thumbnail_width, set_thumbnail_width)


    def get_focus_cell(self):
        return self.real_focus_cell

    def set_focus_cell(self, value):
        if value != self.real_focus_cell:
            self.invalidate_cell(value)
            self.invalidate_cell(self.real_focus_cell)
            self.real_focus_cell = value

    focus_cell = property(get_focus_cell, set_focus_cell)

    def on_key_press(self, widget, event):
        shift = event.state & gtk.gdk.SHIFT_MASK
        ctrl = event.state & gtk.gdk.CONTROL_MASK
        focus_old = self.focus_cell
        keyname = gtk.gdk.keyval_name(event.keyval).lower()
        if keyname == 'down':
            self.focus_cell += self.cells_per_row
        elif keyname == 'left':
            if ctrl and shift:
                self.focus_cell -= self.focus_cell % self.cells_per_row
            else:
                self.focus_cell -= 1
        elif keyname == 'right':
            if ctrl and shift:
                self.focus_cell += self.cells_per_row - (self.focus_cell % self.cells_per_row) - 1
            else:
                self.focus_cell += 1
        elif keyname == 'up':
            self.focus_cell -= self.cells_per_row
        elif keyname == 'home':
            self.focus_cell = 0
        elif keyname == 'end':
            self.focus_cell = len(self.filenames) - 1
        else:
            return False

        self.focus_cell = max(self.focus_cell, 0)
        self.focus_cell = min(self.focus_cell, len(self.filenames)-1)

        if self.focus_cell == focus_old:
            return False

        old_selection = self.selection[:]

        if shift:
            if focus_old != self.focus_cell and focus_old in self.selection and self.focus_cell in self.selection:
                for i in range(min(focus_old, self.focus_cell)+1, max(focus_old, self.focus_cell)+1):
                    if i in self.selection:
                        self.selection.remove(i)
            else:
                for i in range(min(focus_old, self.focus_cell), max(focus_old, self.focus_cell)+1):
                    if not i in self.selection:
                        self.selection.append(i)

        else:
            self.selection = [self.focus_cell]

        self.notify_selection_change(old_selection)

        self.scroll_to(self.focus_cell)
        return True

    def notify_selection_change(self, old_selection):
        for cell in old_selection:
            if not cell in self.selection:
                self.invalidate_cell(cell)
        for cell in self.selection:
            if not cell in old_selection:
                self.invalidate_cell(cell)

    def on_button_press(self, widget, event):
        old_selection = self.selection[:]

        cell_num = self.cell_at_position(event.x, event.y, False)
        if cell_num < 0:
            self.selection = []
            self.notify_selection_change(old_selection, self.selection)
            return False

        if event.type == gtk.gdk.BUTTON_PRESS:
            self.grab_focus()
            ctrl = event.state & gtk.gdk.CONTROL_MASK
            shift = event.state & gtk.gdk.SHIFT_MASK
            if ctrl:
                if cell_num in self.selection:
                    self.selection.remove(cell_num)
                else:
                    self.selection.append(cell_num)
            elif shift:
                for i in range(min(self.focus_cell, cell_num), max(self.focus_cell, cell_num)+1):
                    if not i in self.selection:
                        self.selection.append(i)
            elif not cell_num in self.selection:
                self.selection = [cell_num]

            self.notify_selection_change(old_selection)

            self.queue_resize()

            if event.button != 1:
                return True

            self.focus_cell = cell_num
            return True


    def invalidate_cell(self, order):
        vadjustment = self.get_vadjustment()
        hadjustment = self.get_hadjustment()

        cell_area = self.cell_bounds(order)
        cell_area.width -= 1
        cell_area.height -= 1
        visible = gtk.gdk.Rectangle(int(hadjustment.value), int(vadjustment.value),
                self.allocation.width, self.allocation.height)
        cell_area = cell_area.intersect(visible)
        if self.bin_window and cell_area.width != 0:
            self.bin_window.invalidate_rect(cell_area, False)



    def cell_at_position(self, x, y, crop_visible = True):
        vadjustment = self.get_vadjustment()
        hadjustment = self.get_hadjustment()
        if crop_visible and (
                (y < vadjustment.value or y > vadjustment.value + self.allocation.height) or (
                x < hadjustment.value or x > hadjustment.value + self.allocation.width)):
            return -1

        if x < BORDER_SIZE or x >= BORDER_SIZE + self.cells_per_row * self.cell_width:
            return -1

        if y < BORDER_SIZE or y >= BORDER_SIZE + (len(self.filenames) / self.cells_per_row + 1) * self.cell_height:
            return -1

        column = int((x - BORDER_SIZE) / self.cell_width)
        row = int((y - BORDER_SIZE) / self.cell_height)
        cell_num = column + row * self.cells_per_row

        if cell_num < len(self.filenames):
            return cell_num
        else:
            return -1


    def on_set_scroll_adjustments(self, widget, hadjustment, vadjustment):
        if vadjustment:
            vadjustment.connect('value-changed', self.on_adjustment_value_changed)

    def on_adjustment_value_changed(self, widget, *args):
        self.do_scroll()

    def do_scroll(self):
        vadjustment = self.get_vadjustment()
        hadjustment = self.get_hadjustment()


    def on_size_allocate(self, widget, rectangle):
        self.update_layout(rectangle)
        return True

    def update_layout(self, rectangle = None):
        if rectangle is None:
            rectangle = self.allocation
        #print 'updating layout'
        self.available_width = rectangle[2] - 2 * BORDER_SIZE
        self.cell_width = self.thumbnail_width + 2 * CELL_BORDER_WIDTH
        self.cell_height = self.thumbnail_height + 2 * CELL_BORDER_WIDTH
        self.cells_per_row = max(int(self.available_width / self.cell_width), 1)
        self.cell_width += (self.available_width - self.cells_per_row * self.cell_width) / self.cells_per_row

        if True: # display date
            metrics = self.get_pango_context().get_metrics(self.style.font_desc,
                    pango.Language('en_US'))
            self.cell_height += pango.PIXELS(metrics.get_ascent() + metrics.get_descent())

        self.num_thumbnails = len(self.filenames)
        self.num_rows = int(self.num_thumbnails / self.cells_per_row)
        if self.num_thumbnails % self.cells_per_row:
            self.num_rows += 1

        self.height = int(self.num_rows * self.cell_height + 2 * BORDER_SIZE)

        vadjustment = self.get_vadjustment()
        hadjustment = self.get_hadjustment()
        vadjustment.step_increment = self.cell_height
        x = hadjustment.value
        y = self.height * self.scroll_value
        self.set_size(x, y, self.allocation.width, self.height)

    def set_size(self, x, y, width, height):
        vadjustment = self.get_vadjustment()
        hadjustment = self.get_hadjustment()

        xchange = False
        ychange = False

        hadjustment.upper = max(self.allocation.width, width)
        vadjustment.upper = max(self.allocation.height, height)

        if self.scroll:
            xchange = (hadjustment.value != x)
            ychange = (vadjustment.value != y)
            self.scroll = False

        if self.flags() & gtk.REALIZED:
            self.bin_window.freeze_updates()

        if xchange or ychange:
            if self.flags() & gtk.REALIZED:
                self.bin_window.move_resize(-x, -y, hadjustment.upper, vadjustment.upper)
                vadjustment.value = y
                hadjustment.value = x

        if self.scroll:
            self.scroll = False

        if width != self.allocation.width or height != self.allocation.height:
            gtk.Layout.set_size(self, self.allocation.width, height)

        if xchange or ychange:
            vadjustment.change_value()
            hadjustment.change_value()

        if self.flags() & gtk.REALIZED:
            self.bin_window.thaw_updates()
            self.bin_window.process_updates(True)



    def on_expose_event(self, widget, event):
        self.draw_all_cells(event.area)
        return True

    def get_cell_position(self, cell_num):
        if self.cells_per_row == 0:
            return 0, 0

        row, col = divmod(cell_num, self.cells_per_row)

        x = col * self.cell_width + BORDER_SIZE
        y = row * self.cell_height + BORDER_SIZE

        return x, y

    def draw_all_cells(self, area):
        #print 'draw_all_cells', area.x, area.y, area.width, area.height
        if self.cell_width == 0 or self.cell_height == 0:
            return

        start_cell_column = max((area[0] - BORDER_SIZE)/self.cell_width, 0)
        start_cell_row = max((area[1] - BORDER_SIZE)/self.cell_height, 0)
        start_cell_num = start_cell_column + start_cell_row * self.cells_per_row

        start_cell_x, cell_y = self.get_cell_position(start_cell_num)

        end_cell_column = int(max((area[0] + area[2] - BORDER_SIZE) / self.cell_width, 0))
        end_cell_row = int(max((area[1] + area[3] - BORDER_SIZE) / self.cell_height, 0))

        num_rows = end_cell_row - start_cell_row + 1;
        num_cols = min(end_cell_column - start_cell_column + 1,
                self.cells_per_row - start_cell_column)

        i = 0
        cell_num = start_cell_num
        while i < num_rows and cell_num < len(self.filenames):
            cell_x = start_cell_x
            j = 0
            while j < num_cols and cell_num + j < len(self.filenames):
                #print '  ',
                self.draw_cell(cell_num + j, area)
                cell_x += self.cell_width
                j += 1

            cell_y += self.cell_height
            cell_num += self.cells_per_row
            i += 1


    def cell_bounds(self, cell):
        x, y = self.get_cell_position(cell)
        return gtk.gdk.Rectangle(x, y, self.cell_width, self.cell_height)


    def get_thumbnail_pixbuf(self, thumbnail_num):
        filename = self.filenames[thumbnail_num]
        pixbuf = self.thumbnail_pixbufs.get(filename)
        if pixbuf:
            if pixbuf.get_width() > pixbuf.get_height():
                wx, wy = self.thumbnail_width, self.thumbnail_height
            else:
                r = float(pixbuf.get_height()) / self.thumbnail_height
                wx, wy = int(pixbuf.get_width()/r), self.thumbnail_height
            pixbuf = pixbuf.scale_simple(wx, wy, gtk.gdk.INTERP_BILINEAR)
            return pixbuf

        if not filename in self.priority_loads:
            self.priority_loads.insert(0, filename)
        return self.loading_pixbuf

    def load_thumbnail(self, filename):
        jo = exif.process_file(open(filename))
        self.exif_tags[filename] = jo
        data = jo['JPEGThumbnail']
        loader = gtk.gdk.PixbufLoader()
        loader.write(data, len(data))
        loader.close()
        pixbuf = loader.get_pixbuf()

        if False:
            # scale down thumbnail to match user pref, don't do it so thumbnail
            # size can be changed and scaling will always be done relative to
            # the greatest available size
            wx, wy = self.thumbnail_width, self.thumbnail_height
            if pixbuf.get_width() > pixbuf.get_height():
                wx, wy = self.thumbnail_width, self.thumbnail_height
            else:
                r = float(pixbuf.get_height()) / self.thumbnail_height
                wx, wy = int(pixbuf.get_width()/r), self.thumbnail_height
            pixbuf = pixbuf.scale_simple(wx, wy, gtk.gdk.INTERP_BILINEAR)

        self.thumbnail_pixbufs[filename] = pixbuf
        del pixbuf
        gc.collect()

    def draw_cell(self, thumbnail_num, area):
        #print 'drawing cell', thumbnail_num, self.filenames[thumbnail_num]
        filename = self.filenames[thumbnail_num]

        bounds = self.cell_bounds(thumbnail_num)

        if not bounds.intersect(area).width:
            return

        thumbnail = self.get_thumbnail_pixbuf(thumbnail_num)
        selected = thumbnail_num in self.selection
        if selected:
            if self.flags() & gtk.HAS_FOCUS:
                cell_state = gtk.STATE_SELECTED
            else:
                cell_state = gtk.STATE_ACTIVE
        else:
            cell_state = gtk.STATE_NORMAL
        self.style.paint_flat_box(self.bin_window, cell_state,
                gtk.SHADOW_OUT, area, self, 'ThumbnailsView',
                bounds.x, bounds.y, bounds.width-1, bounds.height-1)

        def inflate(rect, x, y):
            return gtk.gdk.Rectangle(rect.x-x, rect.y-y, rect.width+2*x, rect.height+2*y)

        if self.flags() & gtk.HAS_FOCUS and thumbnail_num == self.focus_cell:
            focus = inflate(bounds, -3, -3)
            self.style.paint_focus(self.bin_window, cell_state,
                    area, self, None, focus.x, focus.y, focus.width, focus.height)

        region = gtk.gdk.Rectangle(0,0,0,0)
        image_bounds = inflate(bounds, -CELL_BORDER_WIDTH, -CELL_BORDER_WIDTH)

        if selected:
            expansion = SELECTION_THICKNESS # XXX; throb expansion (useful?)
        else:
            expansion = 0
        image_bounds = inflate(image_bounds, expansion+1, expansion+1).intersect(area)
        if image_bounds.width:
            def fit(orig_width, orig_height, dest_width, dest_height):
                if orig_width == 0 or orig_height == 0:
                    return 0, 0
                scale = min(dest_width/orig_width, dest_height/orig_height)
                if scale > 1:
                    scale = 1
                fit_width = scale * orig_width
                fit_height = scale * orig_height
                return fit_width, fit_height

            w, h = fit(thumbnail.get_width(), thumbnail.get_height(), 160, 120)
            region = gtk.gdk.Rectangle(
                    bounds.x + (bounds.width - w) / 2,
                    bounds.y + self.thumbnail_height - h + CELL_BORDER_WIDTH,
                    w, h)

            region = inflate(region, expansion, expansion)
            region.width = max(1, region.width)
            region.height = max(1, region.height)
            if region.width != thumbnail.get_width() and region.height != thumbnail.get_height():
                temp_thumbnail = thumbnail.scale_simple(region.width, region.height,
                        gtk.gdk.INTERP_BILINEAR)
            else:
                temp_thumbnail = thumbnail

            region.width = temp_thumbnail.get_width()
            region.height = temp_thumbnail.get_height()

            draw = inflate(region, 1, 1)

            self.style.paint_shadow(self.bin_window, cell_state, gtk.SHADOW_OUT,
                    area, self, 'ThumbnailsView', draw.x, draw.y, draw.width, draw.height)

            draw = region.intersect(area)
            if draw.width:
                self.bin_window.draw_pixbuf(
                        self.style.white_gc,
                        temp_thumbnail,
                        draw.x - region.x,
                        draw.y - region.y,
                        draw.x, draw.y,
                        draw.width, draw.height,
                        gtk.gdk.RGB_DITHER_NONE,
                        draw.x, draw.y)

        layout_bounds = gtk.gdk.Rectangle(0,0,0,0)
        if True: # display dates
            date = self.exif_tags.get(filename, {}).get('EXIF DateTimeOriginal')
            if date:
                layout = pango.Layout(self.get_pango_context())
                layout.set_font_description(self.style.font_desc)
                layout.set_text(str(date))
                layout_bounds.width, layout_bounds.height = layout.get_pixel_size()
                layout_bounds.y = bounds.y + bounds.height - CELL_BORDER_WIDTH - layout_bounds.height + 4
                layout_bounds.x = bounds.x + (bounds.width - layout_bounds.width)/2

                region = layout_bounds.intersect(area)
                if region.width:
                    self.style.paint_layout(self.bin_window, cell_state, True, area, self,
                            'ThumbnailsView', layout_bounds.x, layout_bounds.y, layout)



    def scroll_to(self, cell_num, center = True):
        if not (self.flags() & gtk.REALIZED):
            return
        adjustment = self.get_vadjustment()
        x, y = self.get_cell_position(cell_num)
        if y + self.cell_height > adjustment.upper:
            self.update_layout()

        if center:
            t = y + self.cell_height / 2 - adjustment.page_size / 2
            if t < 0:
                t = 0
            elif t + adjustment.page_size > adjustment.upper:
                t = adjustment.upper - adjustment.page_size
            adjustment.value = t
        else:
            adjustment.value = y




class TestLayoutMgr(gtk.Dialog):
    def __init__(self):
        gtk.Dialog.__init__(self, 'TestLayout', None,
                   gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                   (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                   gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE))
        self.create_gui()

    def create_gui(self):
        self.icon_view = ThumbnailsView()
        sclwin = gtk.ScrolledWindow()
        sclwin.set_policy(gtk.POLICY_NEVER, gtk.POLICY_ALWAYS)
        sclwin.add(self.icon_view)
        self.icon_view.grab_focus()
        self.vbox.pack_start(sclwin)
        hbox = gtk.HBox()
        self.vbox.pack_start(hbox, False, False)
        hscale = gtk.HScale()
        hscale.set_size_request(100, -1)
        hscale.set_property('draw-value', False)
        hscale.set_range(50, 160)
        hscale.connect('value-changed', self.zoom_changed)
        hscale.set_value(160)
        hbox.pack_end(hscale, False, False)
        self.resize(580, 320)
        self.show_all()


    def zoom_changed(self, widget):
        self.icon_view.thumbnail_width = int(widget.get_value())


if __name__ == '__main__':
    if len(sys.argv) == 2:
        path_base = sys.argv[1]
    p = TestLayoutMgr()
    p.run()
    p.destroy()

